@@ -16,12 +16,14 @@ module Clash.Main (defaultMain, defaultMainWithAction) where
 
 -- The official GHC API
 import qualified GHC
-import GHC              (parseTargetFiles,  Ghc, GhcMonad(..), Backend (..),
+import GHC              (parseTargetFiles,  Ghc, GhcMonad(..),
                           LoadHowMuch(..) )
 
+import GHC.Driver.Backend
 import GHC.Driver.CmdLine
 import GHC.Driver.Env
 import GHC.Driver.Errors
+import GHC.Driver.Errors.Types
 import GHC.Driver.Phases
 import GHC.Driver.Session
 import GHC.Driver.Ppr
@@ -40,7 +42,7 @@ import GHC.Platform.Host
 import Clash.GHCi.UI        ( interactiveUI, ghciWelcomeMsg, defaultGhciSettings )
 #endif
 
-import GHC.Runtime.Loader   ( loadFrontendPlugin )
+import GHC.Runtime.Loader   ( loadFrontendPlugin, initializeSessionPlugins )
 
 import GHC.Unit.Env
 import GHC.Unit (UnitId, homeUnitDepends)
@@ -78,12 +80,13 @@ import GHC.Iface.Load
 import GHC.Iface.Recomp.Binary ( fingerprintBinMem )
 
 import GHC.Tc.Utils.Monad      ( initIfaceCheck )
-import System.FilePath
+import GHC.Iface.Errors.Ppr
 
 -- Standard Haskell libraries
 import System.IO
 import System.Environment
 import System.Exit
+import System.FilePath
 import Control.Monad
 import Control.Monad.Trans.Class
 import Control.Monad.Trans.Except (throwE, runExceptT)
@@ -145,10 +148,6 @@ defaultMainWithAction startAction = flip withArgs $ do
     -- 1. extract the -B flag from the args
     argv0 <- getArgs
 
-    -- let (minusB_args, argv1) = partition ("-B" `isPrefixOf`) argv0
-    --     mbMinusB | null minusB_args = Nothing
-    --              | otherwise = Just (drop 2 (last minusB_args))
-
     let argv1 = map (mkGeneralLocated "on the commandline") argv0
     libDir <- ghcLibDir
 
@@ -217,16 +216,16 @@ main' postLoadMode units dflags0 args flagWarnings startAction clashOpts = do
   let dflt_backend = backend dflags0
       (mode, bcknd, link)
          = case postLoadMode of
-               DoInteractive   -> (CompManager, Interpreter,  LinkInMemory)
-               DoEval _        -> (CompManager, Interpreter,  LinkInMemory)
-               DoRun           -> (CompManager, Interpreter,  LinkInMemory)
+               DoInteractive   -> (CompManager, interpreterBackend,  LinkInMemory)
+               DoEval _        -> (CompManager, interpreterBackend,  LinkInMemory)
+               DoRun           -> (CompManager, interpreterBackend,  LinkInMemory)
                DoMake          -> (CompManager, dflt_backend, LinkBinary)
                DoBackpack      -> (CompManager, dflt_backend, LinkBinary)
                DoMkDependHS    -> (MkDepend,    dflt_backend, LinkBinary)
                DoAbiHash       -> (OneShot,     dflt_backend, LinkBinary)
-               DoVHDL          -> (CompManager, NoBackend,     NoLink)
-               DoVerilog       -> (CompManager, NoBackend,     NoLink)
-               DoSystemVerilog -> (CompManager, NoBackend,     NoLink)
+               DoVHDL          -> (CompManager, noBackend,     NoLink)
+               DoVerilog       -> (CompManager, noBackend,     NoLink)
+               DoSystemVerilog -> (CompManager, noBackend,     NoLink)
                _               -> (OneShot,     dflt_backend, LinkBinary)
 
   let dflags1 = dflags0{ ghcMode   = mode,
@@ -253,6 +252,13 @@ main' postLoadMode units dflags0 args flagWarnings startAction clashOpts = do
         where def_ghci_flags = dflags1 `gopt_set` Opt_ImplicitImportQualified
                                        `gopt_set` Opt_IgnoreOptimChanges
                                        `gopt_set` Opt_IgnoreHpcChanges
+                                       -- Setting this by default has the nice effect that
+                                       -- -fno-code and --interactive falls back to interpreter rather than
+                                       -- object code but has little other effect unless you are also using
+                                       -- fat interface files.
+                                       `gopt_set` Opt_UseBytecodeRatherThanObjects
+                                       -- By default enable the debugger by inserting breakpoints
+                                       `gopt_set` Opt_InsertBreakpoints
 
   logger1 <- getLogger
   let logger2 = setLogFlags logger1 (initLogFlags dflags2)
@@ -262,12 +268,13 @@ main' postLoadMode units dflags0 args flagWarnings startAction clashOpts = do
   (dflags3, fileish_args, dynamicFlagWarnings) <-
       GHC.parseDynamicFlags logger2 dflags2 args'
 
-  -- Propagate -Werror to Clash
+-- Propagate -Werror to Clash
   liftIO . modifyIORef' clashOpts $ \opts ->
     opts { opt_werror = EnumSet.member Opt_WarnIsError (generalFlags dflags3) }
 
-  let dflags4 = case bcknd of
-                Interpreter | not (gopt Opt_ExternalInterpreter dflags3) ->
+  let dflags4 = if backendNeedsFullWays bcknd &&
+                   not (gopt Opt_ExternalInterpreter dflags3)
+                then
                     let platform = targetPlatform dflags3
                         dflags3a = dflags3 { targetWays_ = hostFullWays }
                         dflags3b = foldl gopt_set dflags3a
@@ -277,19 +284,20 @@ main' postLoadMode units dflags0 args flagWarnings startAction clashOpts = do
                                  $ concatMap (wayUnsetGeneralFlags platform)
                                              hostFullWays
                     in dflags3c
-                _ ->
+                else
                     dflags3
 
   let logger4 = setLogFlags logger2 (initLogFlags dflags4)
 
   GHC.prettyPrintGhcErrors logger4 $ do
 
-  let flagWarnings' = flagWarnings ++ dynamicFlagWarnings
+  let diag_opts = initDiagOpts dflags4
+  let flagWarnings' = GhcDriverMessage <$> mconcat [warnsToMessages diag_opts flagWarnings, dynamicFlagWarnings]
 
   handleSourceError (\e -> do
        GHC.printException e
        liftIO $ exitWith (ExitFailure 1)) $ do
-         liftIO $ handleFlagWarnings logger4 (initDiagOpts dflags4) flagWarnings'
+         liftIO $ printOrThrowDiagnostics logger4 (initPrintConfig dflags4) diag_opts flagWarnings'
 
   liftIO $ showBanner postLoadMode dflags4
 
@@ -298,16 +306,23 @@ main' postLoadMode units dflags0 args flagWarnings startAction clashOpts = do
   -- we've finished manipulating the DynFlags, update the session
   _ <- GHC.setSessionDynFlags dflags5
   dflags6 <- GHC.getSessionDynFlags
-  hsc_env <- GHC.getSession
+
+  -- Must do this before loading plugins
+  liftIO $ initUniqSupply (initialUnique dflags6) (uniqueIncrement dflags6)
+
+  -- Initialise plugins here because the plugin author might already expect this
+  -- subsequent call to `getLogger` to be affected by a plugin.
+  initializeSessionPlugins
+  hsc_env <- getSession
   logger <- getLogger
 
+
         ---------------- Display configuration -----------
   case verbosity dflags6 of
     v | v == 4 -> liftIO $ dumpUnitsSimple hsc_env
       | v >= 5 -> liftIO $ dumpUnits       hsc_env
       | otherwise -> return ()
 
-  liftIO $ initUniqSupply (initialUnique dflags6) (uniqueIncrement dflags6)
         ---------------- Final sanity checking -----------
   liftIO $ checkOptions postLoadMode dflags6 srcs objs units
 
@@ -418,7 +433,7 @@ checkOptions mode dflags srcs objs units = do
         else do
 
    case mode of
-      StopBefore StopC | backend dflags /= ViaC
+      StopBefore StopC | not (backendGeneratesHc (backend dflags))
         -> throwGhcException $ UsageError $
            "the option -C is only available with an unregisterised GHC"
       StopBefore StopAs | ghcLink dflags == NoLink
@@ -531,7 +546,6 @@ data PostLoadMode
   | DoVerilog               -- ghc --verilog
   | DoSystemVerilog         -- ghc --systemverilog
 
-
 doMkDependHSMode, doMakeMode, doInteractiveMode, doRunMode,
   doAbiHashMode, showUnitsMode, doVHDLMode, doVerilogMode,
   doSystemVerilogMode :: Mode
@@ -678,8 +692,8 @@ mode_flags =
           "LibDir",
           "Global Package DB",
           "C compiler flags",
-          "C compiler link flags",
-          "ld flags"],
+          "C compiler link flags"
+          ],
     let k' = "-print-" ++ map (replaceSpace . toLower) k
         replaceSpace ' ' = '-'
         replaceSpace c   = c
@@ -843,7 +857,7 @@ initMulti unitArgsFiles  = do
     handleSourceError (\e -> do
        GHC.printException e
        liftIO $ exitWith (ExitFailure 1)) $ do
-         liftIO $ handleFlagWarnings logger (initDiagOpts dflags2) warns
+         liftIO $ printOrThrowDiagnostics logger (initPrintConfig dflags2) (initDiagOpts dflags2) (GhcDriverMessage <$> warns)
 
     let (dflags3, srcs, objs) = parseTargetFiles dflags2 (map unLoc fileish_args)
         dflags4 = offsetDynFlags dflags3
@@ -1058,10 +1072,10 @@ showVersion = putStrLn $ concat [ "Clash, version "
 showOptions :: Bool -> IORef ClashOpts -> IO ()
 showOptions isInteractive = putStr . unlines . availableOptions
     where
-      availableOptions opts = concat
-        [ flagsForCompletion isInteractive
-        , map ('-':) (getFlagNames mode_flags)
-        , map ('-':) (getFlagNames (flagsClash opts))
+      availableOptions opts = concat [
+        flagsForCompletion isInteractive,
+        map ('-':) (getFlagNames mode_flags),
+        map ('-':) (getFlagNames (flagsClash opts))
         ]
       getFlagNames opts         = map flagName opts
 
@@ -1170,8 +1184,11 @@ abiHash strs = do
          r <- findImportedModule hsc_env modname NoPkgQual
          case r of
            Found _ m -> return m
-           _error    -> throwGhcException $ CmdLineError $ showSDoc dflags $
-                          cannotFindModule hsc_env modname r
+           _error    ->
+            let opts   = initIfaceMessageOpts dflags
+                err_txt = missingInterfaceErrorDiagnostic opts
+                        $ cannotFindModule hsc_env modname r
+            in throwGhcException . CmdLineError $ showSDoc dflags err_txt
 
   mods <- mapM find_it strs
 
@@ -1218,15 +1235,6 @@ unknownFlagsErr fs = throwGhcException $ UsageError $ concatMap oneError fs
   where
     oneError f =
         "unrecognised flag: " ++ f ++ "\n" ++
-        (case match f (nubSort allNonDeprecatedFlags) of
+        (case flagSuggestions (nubSort allNonDeprecatedFlags) f of
             [] -> ""
             suggs -> "did you mean one of:\n" ++ unlines (map ("  " ++) suggs))
-    -- fixes #11789
-    -- If the flag contains '=',
-    -- this uses both the whole and the left side of '=' for comparing.
-    match f allFlags
-        | elem '=' f =
-              let (flagsWithEq, flagsWithoutEq) = partition (elem '=') allFlags
-                  fName = takeWhile (/= '=') f
-              in (fuzzyMatch f flagsWithEq) ++ (fuzzyMatch fName flagsWithoutEq)
-        | otherwise = fuzzyMatch f allFlags
